__version__ = "2.0.0rc10"

__title__ = "titta"
__description__ = "Titta is a toolbox for using eye trackers from Tobii with Python, specifically offering integration with PsychoPy."
__url__ = "https://github.com/marcus-nystrom/Titta"
__uri__ = __url__
__doc__ = __description__ + " <" + __url__ + ">"

__author__ = "Marcus Nystrom"
__email__ = "marcus.nystrom@humlab.lu.se"

__license__ = "Creative Commons Attribution 4.0 (CC BY 4.0)"
__copyright__ = "Copyright (c) 2019-2023 " + __author__
